//   Async Message-Passing

#include "Channel.h"
#include <iostream.h>

#define pthread_mutexattr_default NULL

template<class T> Channel<T>::Channel()
{
  if (pthread_mutex_init(&s_mux, 
              pthread_mutexattr_default) < 0) 
  {
    cout << "Cannot initialize s_mux." << "\n";
  }
  if (pthread_mutex_init(&r_mux, 
              pthread_mutexattr_default) < 0) 
  {
    cout << "Cannot initialize r_mux." << "\n";
  }
  front = 0; rear = 0;
  empty = new Sema((int)Channel_maxbuf); 
  full = new Sema((int)0); 
}

template<class T> Channel<T>::~Channel()
{
  if ( front != rear ) {
    cout << "Channel not empty." << "\n";
  }
  delete empty;
  delete full;
}

template<class T> void Channel<T>::send(T v)
{
  empty->P();
  if (pthread_mutex_lock(&s_mux) < 0) {
    cout << "send: Cannot lock s_mux." << "\n";
  }
  buffer[rear] = v;
  rear++;
  if (rear >= Channel_maxbuf) {
    rear = 0;
  }
  if (pthread_mutex_unlock(&s_mux) < 0) {
    cout << "send: Cannot unlock s_mux." << "\n";
  }
  full->V();
}

template<class T> void Channel<T>::receive(T* v)
{
  (*full).P();
  if (pthread_mutex_lock(&r_mux) < 0) {
    cout << "receive: Cannot lock r_mux." << "\n";
  }
  *v = buffer[front];
  buffer[front] = 0;
  front++;
  if (front >= Channel_maxbuf) {
    front = 0;
  }
  if (pthread_mutex_unlock(&r_mux) < 0) {
    cout << "receive: Cannot unlock r_mux." << "\n";
  }
  (*empty).V();
}

