// granularity timings

#include <pvm3.h>
#include <pthread.h>
#include <iostream.h>
#include <time.h>
#include "Sema.h"
#include "Channel.cc"

#define maxco	3
#define nloop 1000000

int f(int l) {
  int x = -1;
  return (l*x);
}

void *tf(void ** l) {
  int x = (int)(*l);
  x=-x;
  return (void *)x;
}

static Sema w = Sema(0);
void *tfs(void ** l) {
  int x = (int)(*l);
  x=-x; w.V();
  return (void *)x;
}

static Channel<int> chan = Channel<int>();
void *tfc(void ** l) {
  int x = (int)(*l);
  int v;
  for (;;) {
      chan.receive(&v); 
      if (v < 0) break;
  }
  return (void *)x;
}

static int mytid;
static int msgtag = 1;
void *tfv(void ** l) {
  int x = (int)(*l);
  x=-x; 
  int val[2];
  for (;;) {
    // cout << "+";
      if (pvm_recv(-1, msgtag) < 0) {
         cout << "receive: Cannot receive in tfv. \n" ;
         break;
      }
      if (pvm_upkint(val, 1, 1) < 0) {
         cout << "receive: Cannot get from buffer.\n";
         break;
      }
      x = val[0];
      if (x < 0) break;
  }
  // cout << "+/+";
  return (void *)x;
}


static double factor = -1.0;
void time(int dann, int jetzt, int iter) {
    double it = (double)iter;
    int t = dann-jetzt;
    double f = (double)t/it;
    if (factor <= 0.0) { factor=f; }
    double s = f/factor+0.05;
    cout << "                      " 
         << " ... clocks: " << t 
         << ", factor: " << s 
         << "\n\n";
}


void main()
{
  int iter;
  clock_t ez;
  int es;
  clock_t az = clock();
  int as = az;    // CLOCKS_PER_SEC;
  cout << "start clock: " << as << "\n";

  cout << "starting for loop ... " "\n";
  iter = nloop;
  as = clock();   //CLOCKS_PER_SEC;
  int d = 1;
  int a = 9;
  for (int j=0;j<iter;j++) {
   a = a*d; d=-d;
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);

  cout << "starting for loop with function ... " "\n";
  iter = nloop;
  as = clock();   //CLOCKS_PER_SEC;
  for (int j=0;j<iter;j++) {
   a = a*d; d=f(d);
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);


  cout << "starting for loop with semaphore ... " "\n";
  as = clock();   //CLOCKS_PER_SEC;
  iter = nloop/10;
  Sema s = Sema(0);
  for (int j=0;j<iter;j++) {
   s.V(); a = a*d;  d=-d; s.P();
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);


  cout << "starting for loop with threads ... " "\n";
  as = clock();   //CLOCKS_PER_SEC;
  pthread_t pt;
  int i, o; i=1;
  iter = nloop/300;
  for (int j=0;j<iter;j++) {
      if (pthread_create(&pt, NULL, tf, &i) < 0) {
         cout << "Cannot create thread." << "\n";
      }
      if (pthread_join(pt, (void **)&o) < 0) {
         cout << "Cannot join thread." << "\n";
      }
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);


  cout << "starting for loop with threads and semaphore ... " "\n";
  as = clock();   //CLOCKS_PER_SEC;
  pthread_t pts;
  //int i, o;
  i=1;
  iter = nloop/1000;
  for (int j=0;j<iter;j++) {
      if (pthread_create(&pts, NULL, tfs, &i) < 0) {
         cout << "Cannot create thread." << "\n";
      }
      w.P();
      if (pthread_join(pts, (void **)&o) < 0) {
         cout << "Cannot join thread." << "\n";
      }
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);


  cout << "starting for loop with threads and internal communication ... " "\n";
  as = clock();   //CLOCKS_PER_SEC;
  //int i, o;
  i=1;
  pthread_t ptc;
  if (pthread_create(&ptc, NULL, tfc, &i) < 0) {
      cout << "Cannot create thread." << "\n";
  }
  int v;
  iter = nloop/500;
  for (int j=0;j<iter;j++) {
     if (j<iter-1) { v = 10; } else { v = -1; }
     chan.send(v);
  }
  if (pthread_join(ptc, (void **)&o) < 0) {
     cout << "Cannot join thread." << "\n";
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);


  cout << "starting for loop with threads and pvm communication ... " "\n";
  as = clock();   //CLOCKS_PER_SEC;
  mytid = pvm_mytid();
  cout << "mytaskid " << mytid << "\n";
  //int i, o;
  i=1;
  pthread_t ptv;
  if (pthread_create(&ptv, NULL, tfv, &i) < 0) {
      cout << "Cannot create thread." << "\n";
  }
  int val[2];
  iter = nloop/1000;
  for (int j=0;j<iter;j++) {
     if (pvm_initsend(PvmDataRaw) < 0) {
        cout << "send: Cannot initialize buffer.\n";
        break;
     }
     if (j<iter-1) { val[0] = 10; } else { val[0] = -1; }
     if (pvm_pkint(val, 1, 1) < 0) {
        cout << "send: Cannot put into send-buffer.\n";
        break;
     }
     // cout << "."; 
     if (pvm_send(mytid, msgtag) < 0) {
       cout << "send: Cannot send to task " << mytid << ".\n";
        break;
     }
  }
  if (pthread_join(ptv, (void **)&o) < 0) {
     cout << "Cannot join thread." << "\n";
  }
  es = clock();   //CLOCKS_PER_SEC;
  time(es,as,iter);

  pvm_exit();     // segmentation fault on PIII
  cout << "ended\n";

}

